#
# Copyright (c) 2008-2015 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_form.group module

This module provides groups management to forms. A group is a set of widgets which are displayed
together inside a fieldset; widgets which are not explicitly associated with a group are affected
to a "default" group.
"""

from pyramid.decorator import reify
from zope.interface import implementer
from zope.schema.fieldproperty import FieldProperty

from pyams_form.interfaces import get_form_weight
from pyams_form.interfaces.form import IFormWidgetsGroup, IGroupsBasedForm, IInnerSubForm
from pyams_i18n.interfaces.schema import II18nField


__docformat__ = 'restructuredtext'


@implementer(IFormWidgetsGroup)
class FormWidgetsGroup:
    # pylint: disable=too-many-instance-attributes
    """Form widgets group"""

    form = None
    id = FieldProperty(IFormWidgetsGroup['id'])  # pylint: disable=invalid-name
    widgets = FieldProperty(IFormWidgetsGroup['widgets'])
    bordered = FieldProperty(IFormWidgetsGroup['bordered'])
    fieldset_class = FieldProperty(IFormWidgetsGroup['fieldset_class'])
    legend = FieldProperty(IFormWidgetsGroup['legend'])
    help = FieldProperty(IFormWidgetsGroup['help'])
    _css_class = FieldProperty(IFormWidgetsGroup['css_class'])
    switch = FieldProperty(IFormWidgetsGroup['switch'])
    checkbox_switch = FieldProperty(IFormWidgetsGroup['checkbox_switch'])
    checkbox_field = FieldProperty(IFormWidgetsGroup['checkbox_field'])
    checkbox_mode = FieldProperty(IFormWidgetsGroup['checkbox_mode'])
    display_mode = FieldProperty(IFormWidgetsGroup['display_mode'])
    subforms_legend = FieldProperty(IFormWidgetsGroup['subforms_legend'])

    def __init__(self, form, id, widgets=None, bordered=True, fieldset_class=None, legend=None,
                 help=None, css_class='', switch=False, checkbox_switch=False, checkbox_field=None,
                 checkbox_mode='hide', display_mode='never'):
        # pylint: disable=too-many-arguments,redefined-builtin,invalid-name
        """Form widgets group initialization

        :param form: the form containing the widgets and the group
        :param id: the unique ID (inside the form) of the group
        :param widgets: the widgets names list associated with the group
        :param bordered: if True, the fieldset will have a border
        :param fieldset_class: custom fieldset CSS class
        :param legend: fieldset legend label
        :param help: help text associated with the group
        :param css_class: custom group CSS class
        :param switch: if True, the fieldset will be switchable
        :param checkbox_switch: if True, the switch will be using a checkbox
        :param checkbox_field: if *checkbox_switch* is True, you must define the name of the
            field which will store the checkbox value
        :param checkbox_mode: if set to 'hide', the fieldset content is hidden when the checkbox
            is switched; if set to 'disable', the fieldset contents are only disabled
        :param display_mode: if set to 'never', fieldset content is never displayed; if set to
            'always', fieldset content is always displayed; if set to 'auto', fieldset content is
            displayed when at least one widget value is not default
        """
        assert (not checkbox_switch) or checkbox_field, \
            "You must define checkbox field when using checkbox switch"
        self.form = form
        self.id = id
        self.widgets = widgets or []
        self.bordered = bordered
        self.fieldset_class = fieldset_class
        self.legend = legend
        self.help = help
        self._css_class = css_class
        self.switch = switch
        self.checkbox_switch = checkbox_switch
        self.checkbox_field = checkbox_field
        self.checkbox_mode = checkbox_mode
        self.display_mode = display_mode

    @property
    def css_class(self):
        """Group's CSS class"""
        css_class = self._css_class
        if self.switch:
            if self.checkbox_switch:
                css_class += ' checker'
            else:
                css_class += ' switcher'
        return css_class

    @reify
    def checkbox_widget(self):
        """Get widget associated with checkbox"""
        if self.checkbox_field is None:
            return None
        for widget in self.widgets:
            if widget.field is self.checkbox_field:
                return widget
        return None

    @reify
    def visible(self):
        # pylint: disable=too-many-nested-blocks,too-many-branches
        """Get visible status of group"""
        if self.checkbox_switch:
            widget = self.checkbox_widget
            if self.form.ignoreContext:
                value = widget.field.default
            else:
                context = widget.context
                name = widget.field.getName()
                value = getattr(widget.field.interface(context), name, None)
            return bool(value)
        # else: no checkbox switch
        if self.display_mode == 'never':
            return False
        if (not self.switch) or (self.display_mode == 'always'):
            return True
        # else: no switch or auto mode
        for widget in self.widgets:
            if not widget.ignoreContext:
                field = widget.field
                if self.form.ignoreContext:
                    value = field.default
                else:
                    context = widget.context
                    name = field.getName()
                    value = getattr(field.interface(context), name, None)
                if value and (value != field.default):
                    if II18nField.providedBy(field):
                        for i18n_value in value.values():
                            if i18n_value:
                                return True
                    return True
        return False

    @property
    def visible_widgets(self):
        """Get list of group's visible widgets"""
        for widget in self.widgets:
            if (self.checkbox_field is None) or (widget.field is not self.checkbox_field):
                yield widget

    @property
    def switchable(self):
        """Can fieldset be switched?"""
        return self.switch or self.checkbox_switch

    @property
    def switcher_state(self):
        """Get switcher state"""
        return 'open' if self.visible else 'off'

    @property
    def checker_state(self):
        """Get checker state"""
        return 'on' if self.visible else 'off'

    @reify
    def subforms(self):
        """Get list of subforms associated with this group"""
        registry = self.form.request.registry
        return sorted((adapter[1] for adapter in
                       registry.getAdapters((self.form.context, self.form.request, self),
                                            IInnerSubForm)),
                      key=get_form_weight)

    def get_forms(self):
        """Get list of subforms (including all sub-sub-forms) associated with this group"""
        for form in self.subforms:
            for subform in form.get_forms():
                yield subform

    def update(self):
        """Update all subforms on group update"""
        # pylint: disable=expression-not-assigned
        [subform.update() for subform in self.subforms]

    def update_content(self, content, data):
        """Update content in all subforms"""
        # pylint: disable=expression-not-assigned
        [subform.update_content(content, data) for subform in self.subforms]


def NamedWidgetsGroup(form, id, widgets, names=(), bordered=True, fieldset_class=None,
                      legend=None, help=None, css_class='', switch=False, checkbox_switch=False,
                      checkbox_field=None, checkbox_mode='hide', display_mode='never',
                      factory=FormWidgetsGroup):
    # pylint: disable=too-many-arguments,redefined-builtin,invalid-name
    """Create a widgets group based on widgets names

    :param form: the form containing the widgets and the group
    :param id: the unique ID (inside the form) of the group
    :param widgets: the widgets names list associated with the group
    :param bordered: if True, the fieldset will have a border
    :param fieldset_class: custom fieldset CSS class
    :param legend: fieldset legend label
    :param help: help text associated with the group
    :param css_class: custom group CSS class
    :param switch: if True, the fieldset will be switchable
    :param checkbox_switch: if True, the switch will be using a checkbox
    :param checkbox_field: if *checkbox_switch* is True, you must define the name of the
        field which will store the checkbox value
    :param checkbox_mode: if set to 'hide', the fieldset content is hidden when the checkbox
        is switched; if set to 'disable', the fieldset contents are only disabled
    :param display_mode: if set to 'never', fieldset content is never displayed; if set to
        'always', fieldset content is always displayed; if set to 'auto', fieldset content is
        displayed when at least one widget value is not default
    :param factory: widgets group factory
    """
    if widgets is None:
        form.updateWidgets()
        widgets = form.widgets
    return factory(form, id, [widgets.get(name) for name in names], bordered, fieldset_class,
                   legend, help, css_class, switch, checkbox_switch, checkbox_field,
                   checkbox_mode, display_mode)


@implementer(IGroupsBasedForm)
class GroupsBasedForm:
    """Groups based form

    Should be used as a base class for forms also implementing IForm
    """

    main_group_legend = FieldProperty(IGroupsBasedForm['main_group_legend'])
    main_group_class = FieldProperty(IGroupsBasedForm['main_group_class'])

    def __init__(self):
        self._groups = []

    def add_group(self, group):
        """Add a group to current form's groups"""
        self._groups.append(group)

    @reify
    def groups(self):
        """Get list of groups associated with this form

        If some widgets are associated with any group, a default group is created and located
        at the end of groups list.
        """
        result = self._groups[:]
        others = []
        if self.widgets:  # pylint: disable=no-member
            for widget in self.widgets.values():  # pylint: disable=no-member
                found = False
                for group in result:
                    if widget in group.widgets:
                        found = True
                        break
                if not found:
                    others.append(widget)
            if others:
                result.insert(0, FormWidgetsGroup(self, None, others,
                                                  legend=self.main_group_legend,
                                                  css_class=self.main_group_class))
        return result

    def updateGroups(self):  # pylint: disable=invalid-name
        """Update form groups"""
        # pylint: disable=expression-not-assigned
        [group.update() for group in self.groups]
